# https://raw.github.com/couchrest/couchrest/master/lib/couchrest/design.rb
module CouchRest
  class Design < Document

    # Dispatches to any named list.
    # (using the database where this design doc was saved)
    def list list_name, query={}, &block
      list_on database, list_name, query, &block
    end

    # Dispatches to any named list in a specific database
    def list_on db, list_name, view_name, query = {}, &block
      raise ArgumentError, "List query options must be set as symbols!" if query.keys.find{|k| k.is_a?(String)}
      list_name = list_name.to_s
      list_slug = "#{name}/#{list_name}/#{view_name}"
      # Set the default query options
      query = list_defaults(list_name).merge(query)

      db.list(list_slug, query, &block)
    end

    # Return the hash of default values to include in all queries sent
    # to a list from couchrest.
    def list_defaults(name)
      (self['lists'][name.to_s] && self['lists'][name.to_s]["couchrest-defaults"]) || {}
    end

    # Returns true or false if the view is available.
    def has_list?(name)
      !self['lists'][name.to_s].nil?
    end

    private

    def fetch_list list_name, opts, &block
      database.list(list_name, opts, &block)
    end

  end
end

# https://raw.github.com/couchrest/couchrest/master/lib/couchrest/database.rb
module CouchRest
  class Database
    # == List based queries
    # Query a CouchDB list as defined by a <tt>_design</tt> document. Accepts
    # paramaters as described in http://wiki.apache.org/couchdb/HttpViewApi
    def list(name, params = {}, payload = {}, &block)
      payload['keys'] = params.delete(:keys) if params[:keys]
      #params.delete(:keys)
      # Try recognising the name, otherwise assume already prepared
      list_path = name_to_list_path(name)
      url = CouchRest.paramify_url "#{@root}/#{list_path}", params
      #p [:url, url]
      if block_given?
        if !payload.empty?
          @streamer.post url, payload, &block
        else
          @streamer.get url, &block
        end
      else
        if !payload.empty?
          CouchRest.post url, payload
        else
          CouchRest.get url
        end
      end
    end

    private
    # Convert a simplified list name into a complete list path. If
    # the name already starts with a "_" no alterations will be made.
    def name_to_list_path(name)
      name =~ /^([^_].+?)\/(.*)$/ ? "_design/#{$1}/_list/#{$2}" : name
    end
  end
end

# analogous to https://github.com/couchrest/couchrest_model/blob/master/lib/couchrest/model/designs/view.rb
module CouchRest
  module Model
    module Designs
      class List < View

        attr_accessor :view_name

        def initialize(parent, view_name, new_query = {}, name = nil)
          self.view_name = view_name
          super(parent, new_query, name)
        end

        # == List Execution Methods
        #
        # Request to the CouchDB database using the current query values.
       
        # Return each row wrapped in a ViewRow object. Unlike the raw
        # CouchDB request, this will provide an empty array if there
        # are no results.
        def rows
          return @rows if @rows
          if execute && result['rows']
            @rows ||= result['rows'].map{|v| ViewRow.new(v, model)}
          else 
            [ ]
          end
        end

        def execute
          return self.result if result
          raise "Database must be defined in model or list!" if use_database.nil?

          # Remove the reduce value if its not needed to prevent CouchDB errors
          #query.delete(:reduce) unless can_reduce?
          
          if model.send(view_name.to_sym).can_reduce?          
            query[:reduce] = false if query[:include_docs] # don't reduce if we include_docs
          end

          model.save_design_doc(use_database)

          self.result = model.design_doc.list_on(use_database, name, view_name, query.reject{|k,v| v.nil?})
        end

        class << self
          # Simplified list creation. A new list will be added to the 
          # provided model's design document using the name and options.
          #
          # If the view name starts with "by_" and +:by+ is not provided in 
          # the options, the new list's map method will be interpreted and
          # generated automatically. For example:
          #
          #   List.create(Meeting, "by_date_and_name")
          #
          # Will create a list that searches by the date and name properties. 
          # Explicity setting the attributes to use is possible using the 
          # +:by+ option. For example:
          #
          #   List.create(Meeting, "by_date_and_name", :by => [:date, :firstname, :lastname])
          #
          def create(model, name, function)            
            model.design_doc['lists'] ||= {}
            list = model.design_doc['lists'][name.to_s] = function
            list
          end
        end
      end
    end
  end
end


# https://github.com/couchrest/couchrest_model/blob/master/lib/couchrest/model/designs/view.rb overwritten

class CouchRest::Model::Designs::DesignMapper

  def list(name, function)
    CouchRest::Model::Designs::List.create(model, name, function) if model.auto_update_design_doc
    create_list_method(name)
  end

  def create_list_method(name)
    model.class_eval <<-EOS, __FILE__, __LINE__ + 1
      def self.list_#{name}(view_name, opts = {})
        CouchRest::Model::Designs::List.new(self, view_name, opts, '#{name}')
      end
    EOS
  end

end
